/*
 * File-specific pager functionality.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "file_pager.h"
#include "mapped_file_object_server.h"



/* Initialise a pager for a file with a unique file identifier, file provider,
   mapping flags and a file registry. The provider offers a shared page mapper
   for moderating access to loaded pages. */

FilePager::FilePager(fileid_t fileid, FileProvider *provider, map_flags_t flags)
: Pager(provider->mapper(), flags),
  _provider(provider), fileid(fileid)
{
}

int FilePager::expected_items()
{
    return MappedFileObject_expected_items;
}

ipc_server_handler_type FilePager::handler()
{
    return (ipc_server_handler_type) handle_MappedFileObject;
}



/* Close the pager, removing the provider from the registry if appropriate. */

void FilePager::close()
{
    /* Notify other users of the file. */

    _provider->notify_others(_endpoint, NOTIFY_PEER_CLOSED);

    /* Detach the pager, potentially removing the file provider. */

    _provider->registry()->detach(fileid, _provider);
}



/* File-specific operations. */

long FilePager::flush(offset_t populated_size, offset_t *size)
{
    long err = Pager::flush(populated_size, size);

    if (_resized)
    {
        _provider->notify_others(_endpoint, NOTIFY_CONTENT_AVAILABLE);
        _resized = false;
    }

    return err;
}

long FilePager::resize(offset_t *size)
{
    offset_t old_size = get_data_size();
    long err = Pager::resize(size);

    /* If the size has changed, notify other users, but only after data has been
       flushed. */

    if (old_size < get_data_size())
        _resized = true;

    return err;
}

long FilePager::mmap(offset_t position, offset_t length,
                     offset_t start_visible, offset_t end_visible,
                     offset_t *start_pos, offset_t *end_pos, offset_t *size)
{
    /* Set the limits of the paged region. */

    return Pager::mmap(position, length, start_visible, end_visible, start_pos,
                       end_pos, size);
}



/* Generic pager operations. */

long FilePager::map(offset_t offset, address_t hot_spot, map_flags_t flags, l4_snd_fpage_t *region)
{
    return Pager::map(offset, hot_spot, flags, region);
}



/* Subscribe to notifications. */

long FilePager::subscribe(l4_cap_idx_t notifier, notify_flags_t flags)
{
    /* Readers can subscribe to new data (flush) and file closed events.
       Writers can subscribe to file closed events. */

    return _provider->subscribe(notifier, flags, this, &_endpoint);
}

/* Unsubscribe from notifications. */

long FilePager::unsubscribe()
{
    _provider->unsubscribe(_endpoint, this);
    return L4_EOK;
}

// vim: tabstop=4 expandtab shiftwidth=4
