/*
 * Generic pager functionality.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "dataspace_server.h"
#include "ipc.h"
#include "pager.h"

#include <string.h>



/* Initialise the pager with a page mapper and the given flags controlling
   access to a file. */

Pager::Pager(PageMapper *mapper, map_flags_t flags)
: _start(0), _size(0), _mapper(mapper), _flags(flags),
  _is_masked(false), _start_visible(0), _end_visible(0)
{
}

/* Close the pager. */

void Pager::close()
{
    /* NOTE: Need to deallocate any allocated regions and flexpages. */
}

/* Flush data to the file. */

long Pager::flush(offset_t populated_size, offset_t *size)
{
    _mapper->flush_all(_start, populated_size);

    *size = _mapper->get_data_size();
    return L4_EOK;
}

/* Resize the underlying file. */

long Pager::resize(offset_t *size)
{
    _mapper->set_data_size(*size);

    *size = _mapper->get_data_size();
    return L4_EOK;
}

/* Expose a region of the file. */

long Pager::mmap(offset_t position, offset_t length,
                 offset_t start_visible, offset_t end_visible,
                 offset_t *start_pos, offset_t *end_pos, offset_t *size)
{
    /* Define region characteristics. */

    _start = trunc(position, PAGE_SIZE);
    _size = round(position + length, PAGE_SIZE) - _start;

    /* Return the start and end positions plus populated extent. */

    *start_pos = _start;
    *end_pos = _start + _size;
    *size = _mapper->get_data_size();

    /* Permit masking of mapped regions. */

    if (start_visible || end_visible)
    {
        _start_visible = start_visible;
        _end_visible = end_visible;
        _is_masked = (*start_pos != _start_visible) || (*end_pos != _end_visible);
    }
    else
        _is_masked = false;

    return L4_EOK;
}

/* Map a flexpage corresponding to the dataspace 'offset' involving a 'hot_spot'
   (flexpage offset). */

long Pager::map(offset_t offset, address_t hot_spot, map_flags_t flags,
                l4_snd_fpage_t *region)
{
    offset_t file_offset = _start + offset;
    offset_t max_offset = _start + _size;

    /* Prevent access beyond that defined by the pager.
       NOTE: Permitting executable requests here. This needs to be configured
             when opening the pager or by another means. */

    if (flags & (~(_flags | L4RE_DS_F_X)))
        return -L4_EACCESS;

    Flexpage *issued_flexpage = NULL, *obtained_flexpage = NULL;

    /* Obtain any replicated flexpage.
       NOTE: An additional condition should be introduced to test for the
             relevance of replicated flexpages. */

    if (flags & L4RE_DS_F_W)
        issued_flexpage = _map.find(file_offset);

    /* Without any replicated flexpage, obtain one for the shared, underlying
       file data. */

    if (issued_flexpage == NULL)
    {
        obtained_flexpage = _mapper->get(file_offset, flags);

        /* Determine if the flexpage should be masked. */

        issued_flexpage = get_masked_flexpage(obtained_flexpage);

        /* Determine if the flexpage should be replicated. */

        issued_flexpage = get_replicated_flexpage(issued_flexpage, flags);
    }

    /* Issue the flexpage via the IPC system. */

    long err = ipc_prepare_flexpage(issued_flexpage, file_offset, max_offset,
                                    hot_spot, flags, region);

    if (!err)
        err = complete_Dataspace_map(*region);

    /* After the obtained flexpage is issued, it is queued for future reuse. */

    if (obtained_flexpage != NULL)
        _mapper->queue(obtained_flexpage);

    if (err)
        return err;

    return IPC_MESSAGE_SENT;
}

/* Return the total size of the data. */

offset_t Pager::get_data_size()
{
    return _mapper->get_data_size();
}

/* Detect flexpages with masked content, introducing separate flexpages
   offering masked content from the same region. */

Flexpage *Pager::get_masked_flexpage(Flexpage *flexpage)
{
    /* If not masking content, return the original flexpage. */

    if (!_is_masked)
        return flexpage;

    /* Determine whether the flexpage involves the limits of the visible
       region or is beyond such limits. */

    bool has_start = flexpage->supports_position(_start_visible) &&
                     flexpage->base_offset != _start_visible;
    bool has_end = flexpage->supports_position(_end_visible);
    bool has_zero = (flexpage->base_offset >= _end_visible) ||
                    (flexpage->base_offset + flexpage->size < _start_visible);

    /* Return the original flexpage within the visible limits. */

    if (!has_start && !has_end && !has_zero)
        return flexpage;

    /* Allocate and populate a region in one of the preallocated flexpages for
       masked content. */

    Flexpage &masked = has_start ? _start_flexpage :
                       has_end ? _end_flexpage :
                                 _zero_flexpage;

    allocate_region(flexpage, masked);
    populate_region(flexpage, masked, has_start, has_end);

    /* Associate the preallocated flexpage with the original flexpage. */

    flexpage->associate(&masked);

    return &masked;
}

void Pager::allocate_region(Flexpage *flexpage, Flexpage &masked)
{
    offset_t needed = flexpage->region->size();

    /* Attempt to re-use an existing region. */

    if (masked.region != NULL)
    {
        if (masked.region->size() == needed)
            return;
        else
            _memory.release(masked.region);
    }

    /* Set the region in the flexpage. */

    masked.set_region(_memory.region(needed));
    masked.reset(flexpage->base_offset);
}

void Pager::populate_region(Flexpage *flexpage, Flexpage &masked,
                            bool has_start, bool has_end)
{
    /* Without any limits involved, provide a zero flexpage. */

    if (!has_start && !has_end)
    {
        memset((void *) masked.region->start, 0, masked.size);
        return;
    }

    /* Determine the content limits given start and/or end offsets. */

    offset_t start_offset = has_start ? _start_visible - masked.base_offset : 0;
    offset_t end_offset = has_end ? _end_visible - masked.base_offset : masked.size;

    if (has_start)
        memset((void *) masked.region->start, 0, start_offset);

    memcpy((void *) (masked.region->start + start_offset),
           (const void *) (flexpage->region->start + start_offset),
           end_offset - start_offset);

    if (has_end)
        memset((void *) (masked.region->start + end_offset), 0, masked.size - end_offset);
}

/* Detect flexpages with replicated content, introducing separate flexpages
   offering replicated content from the same region. */

Flexpage *Pager::get_replicated_flexpage(Flexpage *flexpage, map_flags_t flags)
{
    /* NOTE: An additional condition should be introduced to test for the
             relevance of replicated flexpages. For now, detect copy-on-write
             situations. */

    if (_is_masked && (flags & L4RE_DS_F_W))
    {
        Region *region = _memory.region(flexpage->size);
        Flexpage *replicated = new Flexpage(region);

        replicated->reset(flexpage->base_offset);

        memcpy((void *) region->start, (const void *) flexpage->region->start, flexpage->size);
        _map.insert(replicated);
        return replicated;
    }
    else
        return flexpage;
}

// vim: tabstop=4 expandtab shiftwidth=4
